# cython: boundscheck=False
# cython: wraparound=False
# cython: profile=True


import numpy as np
cimport numpy as np

import time
from pybloomfilter import BloomFilter
from bitarray import bitarray
from math import log as log
from math import exp as exp
from math import ceil as ceil
from sets import Set
import config


def test_init():
	assert (config.READ_POSITION_EXPS != [])
	assert (config.mask2_vals != [])
	assert (config.mask1_vals != [])
	print "read length is %d, initialization succeeded" % config.READ_LEN


def get_compressed_bf_params(num, fp, max_size=config.MAX_UNCMP_SIZE):
	cdef int k
	cdef unsigned long m
	k = 0
	m = max_size
	while(m >= max_size):
		try:
			k += 1
			m = (-1.0 * k * int(num)) / log(1 - fp**(1./k))
			print "m is %d, k is %d" % (m, k)
		except ValueError:
			print "on error, fp is %f, k is %d" % (fp, k)
			break
	return (m, k)


def get_bf(reads, fp, filename):
	""" hashes 2bit encoded ints  rep. reads to BF with 
		parameters chosen for minimal compressed file size
		not exceeding max uncompressed size
	"""
	cdef unsigned long long m
	cdef int idx, k
	cdef float num_fake, fp_fake
	
	idx = 0

	numreads = len(reads)#len(read_ba) / (2*READ_LEN)

	(m, k) = get_compressed_bf_params(numreads, fp, config.MAX_UNCMP_SIZE)
	
	# pybloomfilter pkg assumes standard bloom filter
	# since we use compressed bloom filters, we create fake parameters
	num_fake = log(2) * m / float(k)
	fp_fake = (1 - exp(-1 * num_fake * k / float(m) ) )**k
	# print "num_fake is %f, fp_fake is %f, m is %d, k is %d" % (num_fake, fp_fake, m, k)
	bf = BloomFilter(num_fake, fp_fake, filename)
	# bf = BloomFilter(numreads, fp, filename)

	bf.update(reads) # loads reads with one function call
	
	return bf


cdef list get_substitution_hits(object first_query, object bf, 
	list mask1_vals, list mask2_vals, int end_len=10):

	"""given query int, will query bf with all possible 
		substitutions at end for given end length, and return
		the first hit or -1 or if there is none
	"""
	cdef object query, m1, m2, m3
	cdef int j, i

	cdef list query_list, hits 

	hits = []
	query_list = []	
	
	for i in range(end_len):
		m1 = mask1_vals[i] ^ first_query
		m2 = mask2_vals[i] ^ first_query
		m3 = m1 ^ mask2_vals[i]
		query_list.append(m1)
		query_list.append(m2)
		query_list.append(m3)

	for query in query_list:
		if query in bf:
			hits.append(query)

	return hits 

	
cdef list get_strand_scan_reads(object val, unsigned char[:] v, object bf, int end_len):
	cdef list seen_reads, hits, mask1, mask2
	cdef int bytes_per_read, byte_val, byte_pos, read_len, position, i, bits_per_read
	cdef object divisor0, divisor1

	mask1 = config.mask1_vals
	mask2 = config.mask2_vals

	hits = []
	seen_reads = []

	t = time.time()
	read_len = config.READ_LEN
	bytes_per_read = int(ceil(read_len / 4.))
	bits_per_read = 2 * read_len
	byte_pos = 0
	bit_length = val.bit_length()
	divisor0 = <object>2**(<object>bits_per_read-1)
	divisor1 = <object>2**(<object>bits_per_read-2)

	# get next byte, process two bits at a time
	for byte_pos in range(len(v) - (bytes_per_read + 1)):
		
		if val in bf:
			seen_reads.append(val)
		# check 3 subs possibilities at each position for hits
		hits = get_substitution_hits(val, bf, mask1, mask2, end_len)
		seen_reads.extend(hits)
		position = byte_pos + bytes_per_read + 1
		byte_val = v[position] #int(v[position].encode('hex'),16)
		for i in range(4):
			hits = []
			# truncate leftmost 2 bits when needed
			if(val.bit_length() == bits_per_read):
				val = val % divisor0
			if(val.bit_length() == bits_per_read -1):

				val = val % divisor1
			val <<= 2
			# add next two bits of byte
			if (byte_val & 2**(7-2*i)):
				val += 2
			if (byte_val & 2**(7-(2*i+1))):
				val += 1

			if val in bf:
				seen_reads.append(val)
			hits = get_substitution_hits(val, bf, mask1, mask2, end_len)

			seen_reads.extend(hits)

		if (byte_pos % 1000000 == 0):
			print "start is %d, %d reads accepted" % (byte_pos, len(seen_reads))	
			print "%.3f" % (time.time()-t)
	return seen_reads

def query_byte_packed_genome(packed_fasta_file, object bf, int end_len):
	cdef list reads, rev_reads
	cdef object b, b_rc, v, v_rc, val, val_rc
	cdef int i, bits_per_read

	bits_per_read = 2*config.READ_LEN
	print "bf loaded"
	reads = []
	rev_reads = []
	b = bitarray()
	b_rc = bitarray()
	f1=open(packed_fasta_file, 'rb').read()
	b.frombytes(f1)
	print "forward strand loaded"
	v = memoryview(b)
	print "memoryview of forward strand loaded"
	
	# reverse strand of geneme	
	f2=open(packed_fasta_file, 'rb').read()
	b_rc.frombytes(f2)
	b_rc.reverse()
	print "bytes of reverse loaded"
	# for G/C, only need to reverse, 
	# for A/T, also need to invert
	for i in range(0,len(b_rc),2):
		if (b_rc[i]==1 and b_rc[i+1]==1):
			b_rc[i]=0
			b_rc[i+1]=0
		elif (b_rc[i]==0 and b_rc[i+1]==0):
			b_rc[i]=1
			b_rc[i+1]=1
	print "rc fixed"
	v_rc = memoryview(b_rc)
	print "rc memoryview loaded"

	val = <object>int(b[:bits_per_read].tobytes().encode('hex'),16)
	print "val calculated"
	reads = get_strand_scan_reads(val, v, bf, end_len)

	val_rc = <object>int(b_rc[:bits_per_read].tobytes().encode('hex'),16)
	rev_reads = get_strand_scan_reads(val_rc, v_rc, bf, end_len)

	reads.extend(rev_reads)
	
	read_set = set(reads)
	print "%d reads accepted by querying genome" % len(read_set)
	return read_set


def query_bf_on_reads(set reads, object bf):
	cdef list accepts
	cdef set read_set
	cdef object read
	accepts = []
	for read in reads:
		if read in bf:
			accepts.append(read)
	read_set = set(accepts)
	print "%d reads accepted out of %d reads in last set" % (len(read_set), len(reads))
	return read_set



def rec_get_bf(float fp, int end_len, reads_fastq, packed_fasta_file, int iter=2):
	""" hashes 2bit encoded ints rep. reads to BF file,
		then for iter-1 steps hashes FPs accepted by last BF
		to new BF; returns list of FPs relative to last BF
	"""	
	cdef dict FNs
	cdef set read_set, last_set, repeat_set
	cdef int i, read_len, bits_per_read, FN_cnt
	cdef unsigned long long j
	cdef float t
	cdef object val, x

	read_len = config.READ_LEN
	bits_per_read = read_len * 2
	t = time.time()

	FN_cnt = 0
	FNs = {} # FN keys, count vals
	read_set = set([]) # unique reads
	repeat_set = set([]) # repeated reads
	b = bitarray()
	f1=open(reads_fastq, 'rb').read()
	b.frombytes(f1)
	for j in range(0,len(b),bits_per_read):
		val = <object>int(b[j : j + bits_per_read].tobytes().encode('hex'),16)
		if val not in read_set:
			if val not in repeat_set:
				read_set.add(val)
			else:
				FNs[val] += 1 
				FN_cnt += 1
		else:
			read_set.discard(val)
			repeat_set.add(val)
			FNs[val] = 2 
			FN_cnt += 2

	
	# add unique reads to FN, count them in tally
	for val in read_set:
		FNs[val] = 1
		FN_cnt += 1
	print "%.3f seconds to parse known reads" % (time.time() - t)

	for i in range(iter):
		print ""
		print "iteration %d, read set has %d elements" % (i, len(read_set))
		FP_i = []
		T_i = []
		filename = reads_fastq + ".e"+ str(end_len) + ".bf" + str(i)
		if (i==0):
			bf = get_bf(read_set, fp, filename)
			print "approx %d elements in bf, f is %f, %d FNs" % (len(bf), fp, FN_cnt)
			print "querying genome"
			T_i = query_byte_packed_genome(packed_fasta_file, bf, end_len)
			for read in T_i:
				if read not in read_set:
					FP_i.append(read)
				else:
					del FNs[read] # FNs.remove(read)
			last_set = read_set.copy()
			read_set = set(FP_i)
		else:
			bf = get_bf(read_set, fp**i, filename)
			print "approx %d elements in bf, f is %f" % (len(bf), fp)
			print "querying reads"
			T_i = query_bf_on_reads(last_set, bf)
			last_set.clear()
			last_set = read_set.copy()
			read_set = set(T_i)
		del bf
		print "read set for next iteration has %d elements" % len(read_set)

	return (read_set, FNs)

def decode(int end_len, bf_prefix, packed_fasta_file, int iter=4):
	""" uses genome queries to reconstruct read set from BFs
		each read accepted is used to query subsequent BFs
		the set of reads rejected by even numbered BFs + the set
		of reads accepted by the last BF is output
	""" 

	R_rc = [] # reconstructed reads
	q = [] # current query set
	filename = bf_prefix + ".bf" + str(0)
	bf = BloomFilter.open(filename)
	print "querying genome"
	q = query_byte_packed_genome(packed_fasta_file, bf, end_len)
	print "len of q is %d after genome querying" % len(q)

	
	for j in range(1,iter):
		rejects = []
		filename = bf_prefix + ".bf" +str(j)
		bf = BloomFilter.open(filename)
		for read in q:
			if (read not in bf): 
				if (j % 2 == 1):
					R_rc.append(read)	
				rejects.append(read)
		q -= set(rejects)
		print "len of q is %d after querying BF %d" % (len(q), j)
	# output reconstructed reads and left_overs that need to be compared
	# with FP_i list
	return (R_rc, q)
